﻿using System.Net;
using BuildSoft.OscCore;

namespace VRC_Console;

public class OSCCollar
{

    public static void Main(string[] args)
    {
        OSCCollar collar = new OSCCollar();
    }
    
    private OscServer Server { get; init; }
    private OscClient Client { get; init; }
    private float _leashStretch;
    private double _posX, _posY, _verticalMovement, _horizontalMovement;
    private bool _allowMoving;
    private uint _nilSent;
    private DateTime _lastNilMessageSent = DateTime.UnixEpoch;
    private DateTime _lastConsoleOutput = DateTime.UnixEpoch;
    private static readonly TimeSpan ConsoleUpdateInterval = TimeSpan.FromMilliseconds(100);
    private static readonly TimeSpan UpdateFieldsTimeout = TimeSpan.FromMilliseconds(1);
    private static readonly TimeSpan NilMessageMaxTimeout = TimeSpan.FromMilliseconds(400);
    private readonly GPS _gps1 = new GPS(0, -100);
    private readonly GPS _gps2 = new GPS(-86.6, 50);
    private readonly GPS _gps3 = new GPS(86.6, 50);
    private static readonly double calibrationX = -.33;
    private static readonly double calibrationY = -1.1;
    private const double PowerFactor = 1f;
    private string debugValue = "";

    private OSCCollar(string ip = "127.0.0.1", int portReceive = 9001, int portSend = 9000)
    {
        this.Server = new OscServer(portReceive);
        this.Client = new OscClient(ip, portSend);
        Server.TryAddMethod("/avatar/parameters/GPS1", GPS1Handle);
        Server.TryAddMethod("/avatar/parameters/GPS2", GPS2Handle);
        Server.TryAddMethod("/avatar/parameters/GPS3", GPS3Handle);
        Server.TryAddMethod("/avatar/parameters/Leash_Stretch", CollarStretchHandle);
        Server.TryAddMethod("/avatar/parameters/Leash_Toggle", AllowMovingHandle);
        this.Server.Start();

        Thread runningThread = new Thread(RunningThread);
        runningThread.Start();
    }

    private void PrintOutput(double verticalMovement, double horizontalMovement)
    {
        Console.Clear();
        Console.ForegroundColor = ConsoleColor.White;
        Console.WriteLine($"OSC Collar - Status: {(_allowMoving ? "enabled" : "disabled")}");
        Console.WriteLine("==============================");
        Console.WriteLine($"GPS 1:...............{_gps1.Distance:0.000000}");
        Console.WriteLine($"GPS 2:...............{_gps2.Distance:0.000000}");
        Console.WriteLine($"GPS 3:...............{_gps3.Distance:0.000000}");
        Console.WriteLine($"Position X:..........{_posX:0.000000}");
        Console.WriteLine($"Position Y:..........{_posY:0.000000}");
        Console.WriteLine($"Stretch:.............{_leashStretch:0.0000}");
        Console.WriteLine($"Vertical Movement:...{verticalMovement:0.0000}");
        Console.WriteLine($"Horizontal Movement:.{horizontalMovement:0.0000}");
        Console.SetCursorPosition(0, Console.WindowHeight - 2);
        Console.Write($"/input nil sent {_nilSent}");

        if (Console.WindowHeight < 13)
            return;
        Console.SetCursorPosition(40, 5);
        Console.WriteLine("----------+----------");
        for (int i = 1; i < 10; i++)
        {
            if(i == 5)
                continue;
            Console.SetCursorPosition(50,i);
            Console.Write("|");
        }

        int centerX = 50;
        int centerY = 5;
        int consoleX = Convert.ToInt32(Math.Floor(_horizontalMovement * 10));
        int consoleY = Convert.ToInt32(-Math.Floor(_verticalMovement * 4));
        double position = (_verticalMovement * 10) % 10;
        char c = position < 3 ? '.' : position > 7 ? '\'' : 'x';
        Console.SetCursorPosition(centerX + consoleX, centerY + consoleY);
        Console.ForegroundColor = ConsoleColor.Cyan;
        Console.Write(c);
        Console.SetCursorPosition(Console.WindowWidth - 1, Console.WindowHeight - 1);
        
        
        Console.SetCursorPosition(Console.WindowWidth - debugValue.Length - 1, Console.WindowHeight - 2);
        Console.ForegroundColor = ConsoleColor.Gray;
        Console.WriteLine(debugValue);
    }

    private void AllowMovingHandle(OscMessageValues messageValues)
    {
        this._allowMoving = messageValues.ReadBooleanElement(0);
    }

    private void GPS1Handle(OscMessageValues messageValues)
    {
        this._gps1.Distance = messageValues.ReadFloatElement(0) * 100;
    }
    
    private void GPS2Handle(OscMessageValues messageValues)
    {
        this._gps2.Distance = messageValues.ReadFloatElement(0) * 100;
    }
    
    private void GPS3Handle(OscMessageValues messageValues)
    {
        this._gps3.Distance = messageValues.ReadFloatElement(0) * 100;
    }
    
    private void CollarStretchHandle(OscMessageValues messageValues)
    {
        this._leashStretch = messageValues.ReadFloatElement(0);
    }

    private void CalculatePositionFromGPS()
    {
        double a = 2 * _gps2.X - 2 * _gps1.X;
        double b = 2 * _gps2.Y - 2 * _gps1.Y;
        double c = Math.Pow(_gps1.Distance, 2) - Math.Pow(_gps2.Distance, 2) - Math.Pow(_gps1.X, 2) +
            Math.Pow(_gps2.X, 2) - Math.Pow(_gps1.Y, 2) + Math.Pow(_gps2.Y, 2);
        double log = Math.Pow(_gps1.X, 2) +
            Math.Pow(_gps2.X, 2) - Math.Pow(_gps1.Y, 2) + Math.Pow(_gps2.Y, 2);
        double d = 2 * _gps3.X - 2 * _gps2.X;
        //double e = 2 * _gps3.Y - 2 * _gps2.Y;
        double f = Math.Pow(_gps2.Distance, 2) - Math.Pow(_gps3.Distance, 2) - Math.Pow(_gps2.X, 2) +
            Math.Pow(_gps3.X, 2) - Math.Pow(_gps2.Y, 2) + Math.Pow(_gps3.Y, 2);
        
        /*
        _posX = (c * e - f * b) / (e * a - b * d);
        _posY = (c * d - a * f) / (b * d - a * e);*/

        /*
        double a = -173.19;
        double b = 300;
        double c = Math.Pow(_gps1.Distance, 2) - Math.Pow(_gps2.Distance, 2) - 0.44;
        double d = 346.39;
        double f = Math.Pow(_gps2.Distance, 2) - Math.Pow(_gps3.Distance, 2) - 3758.45;*/
        
        _posX = (f / d) + calibrationX;
        _posY = (c * d - a * f) / (b * d) + calibrationY;

        double factor = 1 / Math.Sqrt(Math.Pow(_posX, 2) + Math.Pow(_posY, 2));

        _verticalMovement = _posY * factor * _leashStretch * PowerFactor;
        _horizontalMovement = _posX * factor * -1 * _leashStretch * PowerFactor;
    }

    private void RunningThread()
    {
        while (true)
        {
            this.Server.Update();
            CalculatePositionFromGPS();
            if (_lastNilMessageSent.Add(NilMessageMaxTimeout) < DateTime.Now)
            {
                this.Client.Send("/input/Vertical", 0f);
                this.Client.Send("/input/Horizontal", 0f);
                this._lastNilMessageSent = DateTime.Now;
                _nilSent++;
            }

            if (_allowMoving)
            {
                Thread.Sleep(1);
                if (_leashStretch > 0.1)
                {
                    this.Client.Send("/input/Vertical", Convert.ToSingle(_verticalMovement));
                    this.Client.Send("/input/Horizontal", Convert.ToSingle(_horizontalMovement));
                }
            }

            if (_lastConsoleOutput.Add(ConsoleUpdateInterval) < DateTime.Now)
            {
                PrintOutput(_verticalMovement, _horizontalMovement);
                _lastConsoleOutput = DateTime.Now;
            }
            
            Thread.Sleep(UpdateFieldsTimeout);
        }
    }
}